#
# Python Utils
#
# Author: Carles Mateo
# Creation Date: 2014-11-11 18:18 Barcelona
# Last Update: 2021-04-10 11:15 Ireland
# Description: Class to deal with Python functionalities and compatibility
#


import sys


class PythonUtils:

    def get_python_version(self):
        """
        Returns the version of Python

        Examples:
        sys.version_info(major=2, minor=7, micro=18, releaselevel='final', serial=0)
        sys.version_info(major=3, minor=8, micro=5, releaselevel='final', serial=0)
        :return: s_version, i_major, i_minor, i_micro, s_releaselevel
        """

        # This only works with Python3, so changing to compatible Python2 and Python3
        # i_major = sys.version_info["major"]
        # i_minor = sys.version_info["minor"]
        # i_micro = sys.version_info["micro"]

        i_major = sys.version_info[0]
        i_minor = sys.version_info[1]
        i_micro = sys.version_info[2]

        s_releaselevel = sys.version_info[3]

        s_version = str(i_major) + "." + str(i_minor) + "." + str(i_micro)

        return s_version, i_major, i_minor, i_micro, s_releaselevel

    def is_python_2(self):
        s_version, i_major, i_minor, i_micro, s_releaselevel = self.get_python_version()

        if i_major == 2:
            return True

        return False

    def is_python_3(self):
        s_version, i_major, i_minor, i_micro, s_releaselevel = self.get_python_version()

        if i_major == 3:
            return True

        return False



#
# Date time methods
#
# Author: Carles Mateo
# Creation Date: 2014 Barcelona
# Last Update: 2021-07-01 11:10 Ireland. Added remove_colons
# Description: Class to return Date, datetime, Unix EPOCH timestamp
#

import datetime
import time


class DateTimeUtils:

    def get_unix_epoch(self):
        """
        Will return the EPOCH Time. For convenience is returned as String
        :return: s_now_epoch
        """
        s_now_epoch = str(int(time.time()))

        return s_now_epoch

    def get_unix_epoch_as_int(self):
        """
        Will return the EPOCH Time as Integer.
        :return: i_now_epoch
        """
        i_now_epoch = int(time.time())

        return i_now_epoch

    def get_unix_epoch_as_float(self):
        """
        Will return the EPOCH Time in float with all the decimals.
        :return: f_now_epoch
        """
        f_now_epoch = time.time()

        return f_now_epoch

    def get_datetime(self, b_milliseconds=False, b_remove_spaces_and_colons=False, b_remove_dashes=False):
        """
        Return the datetime with milliseconds in format YYYY-MM-DD HH:MM:SS.xxxxx
        or without milliseconds as YYYY-MM-DD HH:MM:SS"""
        if b_milliseconds is True:
            s_now = str(datetime.datetime.now())
        else:
            s_now = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

        if b_remove_spaces_and_colons is True:
            s_now = s_now.replace(" ", "-")
            s_now = s_now.replace(":", "")

        if b_remove_dashes is True:
            s_now = s_now.replace("-", "")

        return s_now

    def get_datetime_as_14_string(self):
        """
        Returns the datetime in format 202107011444
        :return:
        """

        return self.get_datetime(b_milliseconds=False, b_remove_spaces_and_colons=True, b_remove_dashes=True)



#
# File Utils
#
# Author: Carles Mateo
# Creation Date: 2014-01-01 GMT+1
# Last Update  : 2021-04-11 07:21 Ireland by Carles Mateo
# Description: Class to deal with Files and Directories.
#              Please try to keep it lightweight and without other library dependencies.
#

import os
import glob
import csv
import platform
import sys

# ***********************************************************
# Adding to SysPath so Libraries will find there dependencies
# ***********************************************************
s_path_program = os.path.dirname(__file__)
sys.path.append(s_path_program)
# ***********************************************************

from pythonutils import PythonUtils


class FileUtils:

    o_pythonutils = None

    def test_helper(self):
        pass

    def __init__(self, o_pythonutils=PythonUtils()):
        self.o_pythonutils = o_pythonutils

    def append(self, s_file, s_text):
        """
        This method creates or appends to a file
        :return: boolean
        """
        b_success = True
        try:
            self.test_helper()

            fh = open(s_file, "a")
            fh.write(s_text)
            fh.close()

        except IOError:
            b_success = False

        return b_success

    # Type Hinting and Annotations disabled to have compatibility with Python 3.5.2.
    #     # def create_folder(self, s_folder: str) -> bool:
    def create_folder(self, s_folder):
        """
        Creates a folder
        :param s_folder: Folder to be created
        :return: boolean
        """

        # Commented for Compatibility with Python 2.x
        # b_success: bool = True
        b_success = True
        try:
            os.mkdir(s_folder)
        except IOError:
            b_success = False

        return b_success

    def delete(self, s_file):
        """
        This method deletes a given file
        :param s_file: The file path for the file to delete
        :type s_file: str
        :return: Indicate success of deletion
        :rtype boolean
        """

        b_success = True
        try:
            # This will help with Unit Testing by raisin IOError Exception
            self.test_helper()

            if os.path.exists(s_file):
                os.remove(s_file)
            else:
                b_success = False
        except IOError:
            b_success = False

        return b_success

    def delete_all_with_mask(self, s_mask):
        """
        This method deletes all files matching a given mask. Example mask: '/root/file*.txt'. This mask will delete any
        file in the root directory, starting with 'file' and ending with '.txt'.
        :param s_mask: The mask to use for deletion. Should contain entire file path.
        :type s_mask: str
        :return: Indicate success of deletion
        :rtype boolean
        :return: Number of files successfully deleted
        :rtype integer

        """
        b_success = True
        i_files_deleted = 0
        try:
            self.test_helper()

            for o_file in glob.glob(s_mask):
                os.remove(o_file)
                i_files_deleted = i_files_deleted + 1
        except IOError:
            b_success = False

        return b_success, i_files_deleted

    def get_all_files_or_directories_with_mask(self, s_mask="*"):
        """
        Get all files matching a given mask or the directories.
        Example mask: '/proc/[0-9]*/' will return the directories that start from a number.
        Example mask: '/root/file*.txt'. This mask will return any file in the root directory,
                      which name starts with 'file' and ends in '.txt'.
        :param s_mask: The mask to use for gathering the files. Should contain entire file path.
        :type s_mask: str
        :return: A boolean indicating success, An Array of files matching the mask
        :rtype boolean, list
        """
        a_files = []
        b_success = True
        try:
            self.test_helper()

            for s_filename in glob.glob(s_mask):
                a_files.append(s_filename)
        except IOError:
            b_success = False

        return b_success, a_files

    def get_file_size_in_bytes(self, s_file):

        b_success = False
        i_file_size = 0

        try:
            # This will help with Unit Testing by raisin IOError Exception
            self.test_helper()

            i_file_size = os.path.getsize(s_file)
            b_success = True
        except IOError:
            b_success = False

        return b_success, i_file_size

    def get_real_path(self, s_file):
        """
        Get the canonical path(removes symbolic links) for a specified file
        :param s_file: The file to get the path for
        :return: A boolean indicating success, A string of the real path
        """
        s_path = ""
        try:
            self.test_helper()

            s_path = os.path.realpath(s_file)
            if self.file_exists(s_file):
                b_success = True
            else:
                b_success = False
        except IOError:
            b_success = False

        return b_success, s_path

    def file_exists(self, s_file):
        """
        Check if a file exists
        :param s_file: The file to check
        :return: boolean indicating success
        """
        try:
            self.test_helper()

            b_exists = os.path.isfile(s_file)
        except IOError:
            b_exists = False

        return b_exists

    def folder_exists(self, s_folder):
        """
        Checks if a folder exists
        :param s_folder:
        :return: boolean
        """

        try:
            self.test_helper()

            b_exists = os.path.isdir(s_folder)
        except IOError:
            b_exists = False

        return b_exists

    def list_dir(self, s_dir):
        """
        Get the contents of a directory and return them as a list
        :param s_dir: The directory to list
        :return: A boolean indicating success, A list of the contents
        """
        b_success = True
        a_files = []
        try:
            a_files = os.listdir(s_dir)
        except IOError:
            b_success = False

        return b_success, a_files

    def path_exists(self, s_file_path):
        """
        Check if a file or folder path exists
        :param s_file_path: The file path to check
        :return: boolean indicating success
        """
        try:
            self.test_helper()

            b_exists = os.path.exists(s_file_path)
        except IOError:
            b_exists = False

        return b_exists

    def read_csv(self, s_file):
        """
        Read a CSV file.
        :param s_file:
        :return: boolean, Array
        """
        a_rows = []
        b_error = False

        try:
            with open(s_file, newline='') as o_csv_file:
                o_reader = csv.reader(o_csv_file, delimiter=',')
                for a_row in o_reader:
                    a_rows.append(a_row)
        except:
            b_error = True

        return b_error, a_rows

    def read_file_as_string(self, s_file, s_encoding='utf-8'):
        """
        This method reads the file in text ascii, utf-8 or other encodings.
        Note: Python3 will try to UTF-8 Decode binary and return "UnicodeDecodeError: 'utf-8' codec can't decode byte"
        :param s_file: The file path for the file to read
        :type s_file: str
        :return: b_result: Indicate success reading, s_result: The text of the file
        :rtype boolean, String
        """
        s_result = ""
        b_success = True

        try:
            fh = open(s_file, "r", encoding=s_encoding)
            s_result = fh.read()
            fh.close()

        except IOError:
            b_success = False

        return b_success, s_result

    def read_link_path(self, s_link_path):
        """
        Returns the path to the target gile, given a link
        :param s_link_path: Path to Link
        :return: b_result, s_file_path
        """

        b_success = False

        try:
            s_link_path = os.readlink(s_link_path)
            b_success = True
        except IOError:
            b_success = False
        except OSError:
            b_success = False
        except:
            b_success = False

        return b_success, s_link_path

    def get_only_config_values_from_list(self, a_s_lines):
        """
        Filters out anything that is not of the type key=value or key = value
        :param a_s_lines: the array containing the string lines
        :return: a_s_result
        """
        a_s_result = []

        for s_line in a_s_lines:
            if s_line[0] != "#":
                if "=" in s_line:
                    a_s_result.append(s_line)

        return a_s_result

    def read_config_file_values(self, s_file):
        """
        Reads a config file and returns a dict with key/value
        :param s_file: The file path for the file to read
        :type s_file: str
        :return: b_result, d_s_config_values
        """

        a_s_config = []
        d_s_config_values = {}

        b_success, a_s_result = self.read_file_as_lines(s_file)
        if b_success is True:
            a_s_config = self.get_only_config_values_from_list(a_s_result)

            for s_line in a_s_config:
                a_lines = s_line.split("=")
                s_key = a_lines[0]
                s_value = "=".join(a_lines[1:])
                if len(s_value) > 0 and s_value[-1] == "\n":
                    s_value = s_value[:-1]
                d_s_config_values[s_key] = s_value

        return b_success, d_s_config_values

    def read_file_as_lines(self, s_file, s_encoding='utf-8'):
        """
        This method reads the file in text with the encodinf specified: ascii, utf-8 format.
        Note: Python3 will try to UTF-8 Decode binary and return "UnicodeDecodeError: 'utf-8' codec can't decode byte"
        :param s_file: The file path for the file to read
        :type s_file: str
        :return: b_result: Indicate success reading, a_s_result: The text of the file as an Array of lines
        :rtype boolean, String
        """
        a_s_result = []
        b_success = True

        try:
            self.test_helper()

            with open(s_file, "r", encoding=s_encoding) as fp:
                for cnt, s_line in enumerate(fp):
                    a_s_result.append(s_line)

        except IOError:
            b_success = False

        return b_success, a_s_result

    def read_binary(self, s_file):
        """
        This method reads the file in binary format
        :param s_file: The file path for the file to read
        :type s_file: str
        :return: Indicate success reading
        :rtype boolean
        :return: by_result: byte
        :rtype byte
        """
        by_result = bytes()
        b_success = True

        try:
            self.test_helper()

            fh = open(s_file, "rb")
            by_result = fh.read()
            fh.close()

            # For python2
            by_result = bytearray(by_result)

        except IOError:
            by_result = bytes()
            b_success = False

        except:
            # May have crashed due to not more memory available
            by_result = bytes()
            b_success = False

        return b_success, by_result

    def read_file_with_readlines(self, s_file):
        """
        Read a file and return its contents as an array of lines by using readlines.
        This consideration is important for Memory purposes.
        :param s_file: The file to read
        :return: A list with each line of the file as strings
        """

        a_result = []
        b_success = True

        try:
            self.test_helper()

            fh = open(s_file, "r")
            a_result = fh.readlines()
            fh.close()

        except IOError:
            b_success = False
        except:
            # May have crashed due to not more memory available
            a_result = []
            b_success = False

        return b_success, a_result

    def remove_folder(self, s_folder):
        """
        Remove a folder
        :param s_folder: The folder to remove
        :return: A boolean indicating success
        """

        b_success = True
        try:
            os.rmdir(s_folder)
        except IOError:
            b_success = False

        return b_success

    def rename_file(self, s_old_file, s_new_file):
        """
        Rename a file to a new name
        :param s_old_file: The old file name
        :param s_new_file: The new file name
        :return: A boolean indicating success
        """

        try:
            self.test_helper()

            os.rename(s_old_file, s_new_file)
            b_success = self.file_exists(s_new_file)
        except IOError:
            b_success = False

        return b_success

    def strings(self, s_file):
        """
        This method reads the file as binary and returns only text. Like 'strings' Linux command in text ASCII format.
        :param s_file: The file path for the file to read
        :type s_file: str
        :return: b_result: Indicate success reading, s_result: The text of the file
        :rtype boolean, String
        """
        s_result = ""
        b_success = True

        try:
            self.test_helper()

            # fh = open(s_file, "rb")
            # a_i_result = fh.read()
            # fh.close()

            b_success_read_binary, by_contents = self.read_binary(s_file)

            if b_success_read_binary is False:
                b_success = False
                by_contents = bytes()
            else:
                #for i_ascii in a_i_result:
                for i_ascii in by_contents:
                    if i_ascii > 31 and i_ascii < 127:
                        s_result = s_result + chr(i_ascii)

        except IOError:
            b_success = False

        return b_success, s_result

    def write(self, s_file, s_text):
        """
        This method creates or overwrites a file
        :param s_file: The file path for the file to read
        :type s_file: str
        :param s_text: The text to write
        :type s_text: str
        :return: b_result: Indicate success reading, s_result: The text of the file
        :rtype bool, str
        """
        try:
            fh = open(s_file, "w")
            fh.write(s_text)
            fh.close()

        except IOError:
            return False

        return True

    def write_binary(self, s_file, by_content):
        """
        This method creates or overwrites a file
        :param s_file: The file path for the file to read
        :type s_file: str
        :param by_content: The text to write
        :type by_content: Binary
        :return: b_result: Indicate success reading, by_content: The binary content of the file
        :rtype bool, binary
        """
        try:
            fh = open(s_file, "wb")
            fh.write(by_content)
            fh.close()

        except IOError:
            return False

        return True

    def write_if_changed(self, s_file, s_text):
        """
        This method will check the current file contents against s_text and only write if its different or doesn't exist
        Use on text files, not binary.
        :param s_file: The file to write/update
        :param s_text: The text to write
        :return: boolean: If file was written, str: Error message
        :rtype: bool
        :rtype: str
        """
        # check if the file exists
        b_exists = self.file_exists(s_file)
        # if exists check if it the contents are different
        if b_exists is True:
            b_read_success, s_file_contents = self.read_file_as_string(s_file)
            if b_read_success is True:
                if s_file_contents == s_text:
                    return False, "The contents are the same."
            else:
                return False, "Unable to read File."

        b_write_success = self.write(s_file, s_text)
        if b_write_success is True:
            return True, ""
        else:
            return False, "Failed to write file."

    def get_file_creation_date(self, s_path_to_file):
        """
        Try to get the date that a file was created, falling back to when it was
        last modified if that isn't possible.
        See http://stackoverflow.com/a/39501288/1709587 for explanation.
        :param s_path_to_file: The file
        :return: boolean: If permission was retrieved
        :rtype: bool
        :rtype: int
        """

        i_unix_time = 0
        b_success = True

        if platform.system() == 'Windows':
            i_unix_time = os.path.getctime(s_path_to_file)
        else:
            o_stat = os.stat(s_path_to_file)
            try:
                i_unix_time = o_stat.st_birthtime
            except AttributeError:
                # We're probably on Linux. No easy way to get creation dates here,
                b_success = False
                i_unix_time = -1

        return b_success, i_unix_time

    def get_file_modification_date(self, s_path_to_file):
        """
        Try to get the date that a file was last modified
        :param s_path_to_file: The file
        :return: boolean: If permission was retrieved
        :rtype: bool
        :rtype: int
        """

        i_unix_time = 0
        b_success = True

        if platform.system() == 'Windows':
            i_unix_time = os.path.getmtime(s_path_to_file)
        else:
            o_stat = os.stat(s_path_to_file)
            try:
                i_unix_time = o_stat.st_mtime
            except AttributeError:
                b_success = False

        return b_success, i_unix_time

    def get_own_path(self):
        return os.path.dirname(os.path.abspath(__file__))



#
# Keyboard Utils
#
# Author: Carles Mateo
# Creation Date: 2021-12-07 01:10 Ireland
# Last Update: 2021-12-07 01:10 Ireland
# Description: Class to deal with Python 3 Input from Keyboard
#


class KeyboardUtils:

    def ask_for_string(self, s_text, i_min_length, i_max_length, b_allow_spaces=True, b_allow_underscores=True):
        """
        Ask for a string, and do not leave until criteria is matched.
        :param s_text:
        :param i_min_length:
        :param i_max_length:
        :param b_allow_spaces:
        :param b_allow_underscores:
        :return: String typed by the user
        """
        s_error_msg = "Please enter a text with length between " + str(i_min_length) + " and " + str(i_max_length)
        while True:
            s_answer = input(s_text)
            if b_allow_spaces is False:
                if " " in s_answer:
                    print("Please do not input spaces")
                    continue
            if b_allow_underscores is False:
                if "_" in s_answer:
                    print("Please do not input underscores")
                    continue
            if len(s_answer) < i_min_length or len(s_answer) > i_max_length:
                print(s_error_msg)
                continue

            break

        return s_answer

    def ask_for_valid_integer(self, s_text, i_min, i_max):
        """
        Ask for a Integer, and do not leave until criteria is matched.
        :param s_text:
        :param i_min:
        :param i_max:
        :return:
        """
        s_error_msg = "Please enter a valid value between " + str(i_min) + " and " + str(i_max)
        while True:
            s_answer = input(s_text)
            try:
                i_answer = int(s_answer)
            except ValueError:
                print(s_error_msg)
                continue

            if i_answer < i_min or i_answer > i_max:
                print(s_error_msg)
            else:
                break

        return i_answer



#
# String Utils
#
# Author: Carles Mateo
# Creation Date: 2014-01-01 15:29 IST
# Last Updated:  2021-08-24 20:35 Ireland.
# Description: Class to deal with Files and Directories
#

import os
import glob
import html


class StringUtils:

    def convert_string_to_integer(self, s_amount):
        """
        Convert a string to Integer, checking for errors.
        :param s_amount:
        :return: Boolean Success/False, Integer number if success or 0 otherwise

        """

        b_success = True
        i_value = 0

        if s_amount == "":
            i_value = 0
        else:
            try:
                i_value = int(s_amount)
            except:
                b_success = False
                i_value = 0

        return b_success, i_value

    def convert_to_gb_from_kb(self, s_amount, b_add_units=True):
        """
        Converts units in the format "10KB" or "123456789 KB" or "7000K" to GB with 2 decimal places
        Note: Dividing by 1024 not 1000
        :param s_amount:
        :param b_add_units:
        :return: boolean for success, String in GB with final "GB", float in GB
        """

        s_amount = s_amount.strip()
        f_amount = 0
        b_success = True

        if len(s_amount) < 2:
            if b_add_units is True:
                return False, "0GB", 0
            else:
                return False, "0", 0

        # We will accept "K" not only "KB"
        s_unit_1 = s_amount[-1:].upper()
        if s_unit_1 == "K":
            s_amount = s_amount + "B"

        s_unit = s_amount[-2:].upper()

        if s_unit == "KB":
            s_value = s_amount.replace(' ', '')
            # Remove "KB"
            s_value = s_value[0:-2]
            i_value = int(s_value)
            f_value_mb = i_value/1024.00
            f_value_gb = f_value_mb/1024.00
            #s_value = str(f_value_gb)[0:4]
            s_value = "%.2f" % f_value_gb
            if s_value[-3:] == ".00":
                s_value = s_value[0:-3]
            f_amount = float(s_value)

            if b_add_units is True:
                return b_success, s_value + "GB", f_amount
            else:
                return b_success, s_value, f_amount
        else:
            b_success = False

        # Unknown
        return b_success, s_amount, f_amount

    def convert_bytes_to_best_size(self, i_bytes):
        """
        Converts a number of bytes to the highest values, and adds the units
        :param i_bytes:
        :return: s_biggest_suggested
        """
        s_value_bytes, s_value_kb, s_value_mb, s_value_gb, s_value_tb, s_value_pb, s_biggest_suggested = self.convert_bytes_to_multiple_units(
            i_bytes)

        return s_biggest_suggested

    def convert_bytes_to_multiple_units(self, i_amount):
        s_value_bytes, s_value_kb, s_value_mb, s_value_gb, s_value_tb, s_value_pb, s_biggest_suggested = \
            self.convert_to_multiple_units(str(i_amount), b_add_units=True)

        return s_value_bytes, s_value_kb, s_value_mb, s_value_gb, s_value_tb, s_value_pb, s_biggest_suggested

    def convert_to_multiple_units(self, s_amount, b_add_units=False, i_decimals=2, b_remove_decimals_if_ge_1000=True):
        """
        Getting the bytes or Kbytes, we return multiple units.
        :param s_amount:
        :param b_add_units: If we suffix the units, like GiB
        :param i_decimals: Number of decimal positions
        :param b_remove_decimals_if_ge_1000: Will remove decimals part if the unit is greater or equal to 1000
        :return: String in Bytes, String in Kbytes, String in Mbytes, String in Gbytes, String in Tbytes,
                 str in Pbytes, biggest suggested unit
        """
        # @TODO: Finish this to support other units as input
        # Init values
        s_value_bytes = "0"
        i_value_bytes = 0
        s_value_kb = "0"
        i_value_kb = 0
        s_value_mb = "0"
        f_value_mb = 0
        s_value_gb = "0"
        f_value_gb = 0
        s_value_tb = "0"
        f_value_tb = 0
        s_value_pb = "0"
        f_value_pb = 0

        s_biggest_suggested = ""

        s_amount = s_amount.strip()

        s_mask = "{:." + str(i_decimals) + "f}"  # "{:.2f}"

        # Check if it has units, like 10KB
        if s_amount.isdigit() is not True:
            s_unit = s_amount[-2:].upper()
            s_unit1 = s_amount[-1:].upper()
            s_unit3 = s_amount[-3:].upper()

            if s_unit1 == "K" or s_unit == "KB" or s_unit3 == "KIB":
                s_value = s_amount.replace(' ', '')
                # Remove "KB"
                if s_unit == 'KB':
                    s_value = s_value[0:-2]
                elif s_unit3 == 'KIB':
                    s_value = s_value[0:-3]
                elif s_unit1 == "K":
                    s_value = s_value[0:-1]
                i_value = int(s_value)
                i_value_bytes = i_value * 1024
                # Convert to Bytes
                s_amount = str(i_value_bytes)

        if s_amount.isdigit() is True:
            # Is Bytes
            i_value = int(s_amount)
            i_value_bytes = i_value
            s_value_bytes = str(i_value_bytes)
            i_value_kb = i_value_bytes / 1024.00
            s_value_kb = s_mask.format(i_value_kb)
            f_value_mb = i_value_bytes / 1024 / 1024.00
            s_value_mb = s_mask.format(f_value_mb)
            f_value_gb = i_value_bytes / 1024 / 1024 / 1024.00
            s_value_gb = s_mask.format(f_value_gb)
            f_value_tb = i_value_bytes / 1024 / 1024 / 1024 / 1024.00
            s_value_tb = s_mask.format(f_value_tb)
            f_value_pb = i_value_bytes / 1024 / 1024 / 1024 / 1024 / 1024.00
            s_value_pb = s_mask.format(f_value_pb)

        if b_remove_decimals_if_ge_1000 is True:
            if i_value_kb >= 1000:
                i_value_kb = int(i_value_kb)
                s_value_kb = str(i_value_kb)
            if f_value_mb >= 1000:
                f_value_mb = int(f_value_mb)
                s_value_mb = str(f_value_mb)
            if f_value_gb >= 1000:
                f_value_gb = int(f_value_gb)
                s_value_gb = str(f_value_gb)
            if f_value_tb >= 1000:
                f_value_tb = int(f_value_tb)
                s_value_tb = str(f_value_tb)
            if f_value_pb >= 1000:
                f_value_pb = int(f_value_pb)
                s_value_pb = str(f_value_pb)
        else:
            s_value_mb = s_mask.format(f_value_mb)
            s_value_gb = s_mask.format(f_value_gb)
            s_value_tb = s_mask.format(f_value_tb)
            s_value_pb = s_mask.format(f_value_pb)

        if b_add_units is True:
            if i_value_bytes > 1 or i_value_bytes == 0:
                s_value_bytes = s_value_bytes + "Bytes"
            elif i_value_bytes == 1:
                s_value_bytes = s_value_bytes + "Byte"

            s_value_kb = s_value_kb + "KB"
            s_value_mb = s_value_mb + "MB"
            s_value_gb = s_value_gb + "GB"
            s_value_tb = s_value_tb + "TB"
            s_value_pb = s_value_pb + "PB"

        if f_value_pb >= 1:
            s_biggest_suggested = s_value_pb
        elif f_value_tb >= 1:
            s_biggest_suggested = s_value_tb
        elif f_value_gb >= 1:
            s_biggest_suggested = s_value_gb
        elif f_value_mb >= 1:
            s_biggest_suggested = s_value_mb
        elif i_value_kb >= 1:
            s_biggest_suggested = s_value_kb
        else:
            s_biggest_suggested = s_value_bytes

        return s_value_bytes, s_value_kb, s_value_mb, s_value_gb, s_value_tb, s_value_pb, s_biggest_suggested

    def format_float_to_string(self, f_number, i_decimal_positions=2):
        s_format = "%." + str(i_decimal_positions) + "f"

        s_formatted = (s_format % f_number)

        return s_formatted

    def get_dict_value(self, s_key, d_dict, m_default=""):
        """
        Returns the value of the dictionary for key or default. Same as Dict.get
        :param s_key:
        :param d_dict:
        :param m_default:
        :return: Boolean found or not, value
        """
        if s_key in d_dict:
            return True, d_dict[s_key]

        return False, m_default

    def add_trailing_slash(self, s_path):
        """
        Adds / if not found at the end
        :param s_path:
        :return: String
        """

        if s_path == "":
            s_path = "/"

        if s_path[-1] != "/":
            s_path = s_path + "/"

        return s_path

    def format_string_to_fixed_length(self, s_text, i_length=10, s_align_mode="L", b_truncate_excess=True):
        """
        Formats a string keeping the max characters and aligning
        :param b_truncate_excess:
        :param i_length:
        :param s_text:
        :param s_align_mode:
        :return: str
        """

        s_length = str(i_length)

        if b_truncate_excess is True:
            s_text = s_text[0:i_length]

        if s_align_mode == "L":
            s_format = '{0: <' + s_length + '}'
        else:
            # s_align_mode == "R":
            s_format = '{0: >' + s_length + '}'

        s_output = s_format.format(s_text)

        return s_output

    def replace_and_format_multiple_strings(self, s_original_text, a_h_s_to_render):
        """
        Replaces and formats multiple strings based on a_h_s_to_render.
        Example:
        a_h_s_format = [{"s_replace_from": "#DATA_SOURCE#",
                                 "s_replace_to": s_id_data_source,
                                 "i_length": i_length_field_datasource,
                                 "s_align_mode": "L",
                                 "b_truncate_excess": True
                        },
                        {"s_replace_from": "#ENABLED#",
                         "s_replace_to": "True",
                         "i_length": i_length_field_enabled,
                         "s_align_mode": "R",
                         "b_truncate_excess": True
                         }]
        :param s_original_text:
        :param a_h_s_to_render:
        :return:
        """

        s_output = s_original_text

        for h_s_to_render in a_h_s_to_render:
            s_replace_from = h_s_to_render["s_replace_from"]
            s_replace_to = h_s_to_render["s_replace_to"]
            i_length = h_s_to_render["i_length"]
            s_align_mode = h_s_to_render["s_align_mode"]
            b_truncate_excess = h_s_to_render["b_truncate_excess"]

            s_field_aligned = self.format_string_to_fixed_length(s_replace_to,
                                                                 i_length=i_length,
                                                                 s_align_mode=s_align_mode,
                                                                 b_truncate_excess=b_truncate_excess)

            s_output = s_output.replace(s_replace_from, s_field_aligned)

        return s_output

    def convert_integer_to_string_thousands(self, i_number, s_thousand_sep=","):
        """
        Puts thousand separator
        :param i_number:
        :param s_thousand_sep:
        :return: s_number_formatted
        """

        s_number = str(i_number)
        s_number_formatted = ""
        i_pos = 0
        for i_index in range(len(s_number)-1, -1, -1):
            i_pos = i_pos + 1
            if i_pos % 3 == 1 and i_pos > 1 and s_number[i_index] != "-":
                s_number_formatted = s_number[i_index] + s_thousand_sep + s_number_formatted
            else:
                s_number_formatted = s_number[i_index] + s_number_formatted

        return s_number_formatted

    def get_bytes_per_second(self, i_bytes, f_seconds):
        """
        Returns the speed in bytes/seconds
        :param i_bytes:
        :param f_seconds:
        :return: i_speed_bytes_per_second, s_biggest_suggested
        """

        if f_seconds == 0:
            # If 0 seconds were sent as interval, we consider that it's 1
            # This way we avoid a Division by zero
            f_seconds = 1

        i_speed_bytes_per_second = int(i_bytes / f_seconds)
        s_value_bytes, s_value_kb, s_value_mb, s_value_gb, s_value_tb, s_value_pb, s_biggest_suggested = \
            self.convert_bytes_to_multiple_units(i_speed_bytes_per_second)

        return i_speed_bytes_per_second, s_biggest_suggested

    def get_percent(self, i_partial, i_total, b_add_percent=True):
        """
        Gets the percent from two integers.
        Is designed to work with Python 3 and Python 2.
        Note: Corner case i_total == 0 considered as 100%.
        """

        if i_total == 0:
            # We avoid division by zero
            f_percent_free = 0
            i_percent_free = 0
            s_percent_decimals = "100"
            s_percent_no_decimals = "100"
        else:
            f_percent_free = (100.0 * i_partial) / i_total
            i_percent_free = int(f_percent_free)
            s_percent_no_decimals = str(i_percent_free)
            s_percent_decimals = str(round(f_percent_free, 2))

        if b_add_percent is True:
            s_percent_no_decimals = s_percent_no_decimals + "%"
            s_percent_decimals = s_percent_decimals + "%"

        return f_percent_free, i_percent_free, s_percent_decimals, s_percent_no_decimals

    def get_time_and_bytes_per_second(self, i_bytes, f_time_start, f_time_finish):
        """
        Calculates the time between start and end, and the bytes per second, and the best unit for it
        :param i_bytes:
        :param f_time_start:
        :param f_time_finish:
        :return: f_time, s_time, i_bytes_per_second, s_best_unit
        """

        f_time = f_time_finish - f_time_start
        s_time = str(round(f_time, 2))

        i_bytes_per_second, s_best_unit = self.get_bytes_per_second(i_bytes, f_time)

        return f_time, s_time, i_bytes_per_second, s_best_unit

    def html_escape(self, s_text, b_quote=True):
        """
        Returns HTML safe, with escaped characters like <> and &
        :param s_text: Text to escape-HTML from
        :return: String
        """

        s_result = html.escape(s_text, quote=b_quote)

        return s_result



#
# Class to work with subprocesses
#
# Author: Carles Mateo
# Creation Date: 2014
#
# v.0.1.4 - 2020-10-04 - Carles Mateo
# v.0.1.5 - 2020-10-09 - Carles Mateo
# v.0.1.6 - 2020-10-11 - Carles Mateo
#

import subprocess


class SubProcessUtils:

    def execute_command_for_output(self, s_command, b_shell=True, b_convert_to_ascii=True, b_convert_to_utf8=False):
        """
        Executes Command and returns the Error Code, the Standard Output STDOUT and Standard Error STDERR
        """

        s_output = ""
        s_error = ""

        try:
            # Note: In Python3 this will return a binary
            process = subprocess.Popen([s_command],
                                       shell=b_shell,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            s_output, s_error = process.communicate()
            i_error_code = process.returncode

        except:
            i_error_code = 99

        try:
            if b_convert_to_ascii is True:
                s_output = s_output.decode('ascii')
                s_error = s_error.decode('ascii')
            if b_convert_to_utf8 is True:
                s_output = s_output.decode('utf8')
                s_error = s_error.decode('utf8')

        except:
            i_error_code = 128

        return i_error_code, s_output, s_error

    def execute_command_with_no_pipe_and_no_stderr(self, s_command, b_shell=True, b_convert_to_ascii=True, b_convert_to_utf8=False):
        """
        Execute without locking pipe and without stderr
        """

        # Using stderr = subprocess.STDOUT would provoke locking and expiring the timeout
        s_output = ""
        s_exception = ""
        try:
            s_output = subprocess.check_output([s_command], shell=b_shell)
            if b_convert_to_ascii is True:
                s_output = s_output.decode('ascii')
            if b_convert_to_utf8 is True:
                s_output = s_output.decode('utf8')

        except:
            s_output = ""

        return s_output



from datetimeutils import DateTimeUtils
from keyboardutils import KeyboardUtils
from menuutils import MenuUtils
from fileutils import FileUtils
from pythonutils import PythonUtils
from stringutils import StringUtils
from subprocessutils import SubProcessUtils


class CarlesLibsDemo:
    """
    A demo apllication for carleslibs from Carles Mateo
    https://blog.carlesmateo.com/carleslibs
    """

    def __init__(self, o_datetime, o_keyboard, o_menu, o_python, o_file):
        self.o_datetime = o_datetime
        self.o_keyboard = o_keyboard
        self.o_menu = o_menu
        self.o_python = o_python
        self.o_file = o_file

    def main(self):
        s_version, i_major, i_minor, i_micro, s_releaselevel = self.o_python.get_python_version()
        print("Python Version:", s_version)
        if self.o_python.is_python_2() is True:
            print("Warning! Python 2 reached End of Life")
        print()

        a_menu = [("See all the files in the local directory", False, self, "list_files"),
                  ("See files with a specific mask", False, self, "list_files_asking"),
                  ("Print current time", False, self, "print_current_time")]

        self.o_menu.run_menu(s_title="Carleslibs Demo Main Menu", a_menu=a_menu, b_admin_user=False, s_msg_select="Select menu option:", s_msg_return="Return (or CTRL + C to exit the program)")

    def print_current_time(self):
        print("Current time:", self.o_datetime.get_datetime(), "-", self.o_datetime.get_unix_epoch())

    def list_files_asking(self):

        s_mask = self.o_keyboard.ask_for_string("Mask (Ie: *.py):", i_min_length=0, i_max_length=100, b_allow_spaces=False, b_allow_underscores=True)
        if s_mask == "":
            print("No mask selected.")
            return

        self.list_files(s_mask)

    def list_files(self, s_mask="*"):
        b_success, a_files = self.o_file.get_all_files_or_directories_with_mask(s_mask)

        if b_success is False:
            print("There has been an error")
            return

        print(self.o_menu.get_nice_title("Files with mask " + s_mask, s_underline_char="-"))
        print()

        print("Total files:", len(a_files))
        print()

        if len(a_files) > 0:
            for s_file in a_files:
                print(s_file)

            print()


if __name__ == "__main__":
    o_datetime = DateTimeUtils()
    o_keyboard = KeyboardUtils()
    o_menu = MenuUtils()
    o_python = PythonUtils()
    o_file = FileUtils(o_pythonutils=o_python)

    o_demo = CarlesLibsDemo(o_datetime, o_keyboard, o_menu, o_python, o_file)
    try:
        o_demo.main()
    except KeyboardInterrupt:
        print("CTRL + C captured. Exitting")



from keyboardutils import KeyboardUtils


class MenuUtils:

    def __init__(self):

        self.o_keyboard = KeyboardUtils()

    def run_menu(self, s_title, a_menu, b_admin_user=False, s_msg_select="Select menu option:", s_msg_return="Return (or CTRL + C to exit the program)"):
        """
        Prints a menu and offers a selection, then executes the class and method indicated in 3th and 4th positions of the tuple
        for that line.
        :param s_title: Title for the menu
        :param a_menu: Menu as an array of tuples, with tuples like ("Line in the menu", b_is_visible_for_admin_only, o_class, "method_name_in_object_o_class")
        :param b_admin_user: If the user for who we are rendering the menu is admin (so it can see the options with b_is_visible_for_admin_only set to True)
        :param s_msg_select:
        :param s_msg_return:
        :return: None
        """

        # This will loop until 0 - Return is selected
        while True:
            a_menu_mapping = []
            i_menu_index = 0
            # Map the function to the index of the menu
            # 0 will be always go back
            a_menu_mapping.append("")

            print(s_title)
            print("=" * len(s_title))
            print()

            for t_menu_item in a_menu:
                s_title = t_menu_item[0]
                b_admin = t_menu_item[1]
                o_object = t_menu_item[2]
                s_method = t_menu_item[3]
                if (b_admin is True and b_admin_user is True) or (b_admin is False):
                    i_menu_index += 1
                    a_menu_mapping.append((o_object, s_method))
                    print(str(i_menu_index) + ". " + s_title)

            print("0. " + s_msg_return)

            i_option = self.o_keyboard.ask_for_valid_integer(s_msg_select, 0, i_menu_index)
            # Execute the command
            if i_option > 0:
                o_object = a_menu_mapping[i_option][0]
                s_method = a_menu_mapping[i_option][1]
                result = getattr(o_object, s_method)()

            if i_option == 0:
                return

    def run_menu_for_selection(self, a_menu, s_msg_return="Return (or CTRL + C to exit the program)"):
        """
        Print the menu, offer selection, return the string associated to the index
        :param a_menu:
        :return: String, String: Data of first selection, Data optional
        """

        i_justify_spaces = 1
        if len(a_menu) > 99:
            i_justify_spaces = 3
        elif len(a_menu) > 9:
            i_justify_spaces = 2

        i_menu_index = 0
        while True:
            for t_menu_item in a_menu:
                s_title = t_menu_item[0]
                s_value = t_menu_item[1]
                i_menu_index += 1
                # We reserve spaces as we can have up to 20 VMs running and many more stopped
                # We want to keep the alignment
                print(str(i_menu_index).rjust(i_justify_spaces) + ". " + s_title)

            print(s_msg_return)

            i_option = self.o_keyboard.ask_for_valid_integer("Select:", i_min=0, i_max=i_menu_index)
            # Execute the command
            if i_option > 0:
                s_value = a_menu[i_option-1][1]
                s_value_optional = ""
                if len(a_menu[i_option-1]) > 2:
                    s_value_optional = a_menu[i_option-1][2]
                return s_value, s_value_optional

            if i_option == 0:
                return "", ""

    def get_nice_title(self, s_text, s_underline_char="="):
        """
        Will return the s_text with underline and a blank line
        :param s_text:
        :param s_underline_char:
        :return: string: s_text_formatted
        """

        s_text_formatted = "\n"
        s_text_formatted += s_text + "\n" + s_underline_char * len(s_text) + "\n"

        s_text_formatted += "\n"

        return s_text_formatted



def read_file_as_lines(s_file, s_encoding='utf-8'):
    """
    This method reads the file in text with the encodinf specified: ascii, utf-8 format.
    Note: Python3 will try to UTF-8 Decode binary and return "UnicodeDecodeError: 'utf-8' codec can't decode byte"
    :param s_file: The file path for the file to read
    :type s_file: str
    :return: b_result: Indicate success reading, a_s_result: The text of the file as an Array of lines
    :rtype boolean, String
    """
    a_s_result = []
    b_success = True

    try:

        with open(s_file, "r", encoding=s_encoding) as fp:
            for cnt, s_line in enumerate(fp):
                a_s_result.append(s_line)

    except FileNotFoundError:
        print("File not found")
    except IOError:
        print("IOError")
        b_success = False

    return b_success, a_s_result

b_success, a_s_result = read_file_as_lines("nonexisting")
print(b_success)